/**********
TITLE :         4 bpp sprite demo.
AUTHOR :        XL2
DESCRIPTION :   This is a little demo to show how to use a color lookup table and 16 color bank, using JO Engine, but could be adapted for SGL only easily.
                Use the included windows converter app by XL2 to create a binary file or a .h file with a 16 colors palette
CAUTION :       Make sure your image has 16 colors or less, else the Windows app will just compress the palette by removing the LSB until it fits.
                Also, make sure your images don't use RLE compression and try to use 32 bpp data with alpha channel for the transparency.
                Be also careful : if your texture has transparent pixels, make sure enable it (SPenb) in your sprite attributes, else disable it (SPdis).
VERSION :       0.2 (2018-01-06)
**********/

#include <jo/jo.h>

#define LWRAM 0x00200000
/*********************
Color lookup tables data
*********************/
#define     LUTaddr             (0x25C7D960)
#define     returnLUTaddr(n)    (LUTaddr + (0x20 * n))
#define     LUTidx(n)           ((0x7D960 + (0x20 * n))/8)
#define     CRAM_Base           (0x25f00200)
#define     returnCRAMaddr(n)   (CRAM_Base + (0x20 * n))
#define     LUTcramIdx(n)       ((256 + (16 * n)))

/**Only used for the C file**/
typedef struct
{
    Uint16 width;
    Uint16 height;
    Uint8 * palette_id;  //The palette index for the pixels
    Uint16 * palette; //The colors (16 colors)
} img_4bpp;

#include "sky.h"
/*******END*************/



/**Simple function to draw the sprites with Color Lookup tables**/
void render_CLUT_sprite(unsigned int id, int x, int y, int z)
{
    FIXED pos[XYZS];    pos[X]=toFIXED(x); pos[Y]=toFIXED(y); pos[Z]=toFIXED(z); pos[S]=65536;
    SPR_ATTR spr_attributes = SPR_ATTRIBUTE(id, LUTidx(id), No_Gouraud, CL16Look | ECdis | SPdis | HSSon ,  sprNoflip  );
    slPutSprite(pos , &spr_attributes , 0) ;
}


/**Simple function to draw the sprites with Color Bank from Color RAM**/
void render_CRAM_sprite(unsigned int id, int x, int y, int z)
{
    FIXED pos[XYZS];    pos[X]=toFIXED(x); pos[Y]=toFIXED(y); pos[Z]=toFIXED(z); pos[S]=65536;
    SPR_ATTR spr_attributes = SPR_ATTRIBUTE(id, LUTcramIdx(id), No_Gouraud, CL16Bnk | ECdis | SPdis | HSSon , sprNoflip  );
    slPutSprite(pos, &spr_attributes , 0) ;
}

/***Rendering callback function**/
void			    my_draw(void)
{
    render_CLUT_sprite(0, -48, -20, 160);
    render_CLUT_sprite(1, 48, -20, 150);
    render_CRAM_sprite(2, -24, 48, 110);
    render_CRAM_sprite(3, 150, 100, 400);  //This image got auto-compressed to 16 colors, which is why it's not optimal
}


/***************************
This sets palette
***************************/
void	set_palette(Uint16 * palette, Uint16 TextureId)
{
    //For SGL only use slDMACopy instead : slDMACopy( (void*)palette, (void*)adr, sizeof(palette) );
    jo_dma_copy(palette, (void*)(returnLUTaddr(TextureId)), sizeof(Uint16)*16);

    //Only if you want to use the VDP2 CRAM for your sprite, else skip this
    jo_palette_to_cram(palette, (void*)(returnCRAMaddr(TextureId)), 16);
}


/**************************
Load the image using the .h file
Using the jo_img_8bits for a 4 bits image as an easy workaround (16 bits also works, see below)
**************************/
void			    load_4bits_image(void)
{
    jo_img_8bits    img;
    jo_texture_definition   *texture;
    img.data = sky_DATA;
    img.width = JO_DIV_BY_2(sky.width);  //Since the image is 4 bits per pixel, we divide the width by 2 to fit each 2 pixels in 1 Byte
    img.height = sky.height;
    int id = jo_sprite_add_8bits_image(&img);  //Adds the image in memory
    texture=&__jo_sprite_def[id];
    texture->width=sky.width;  //Ghetto technique for compatibility with Jo Engine, but trying to replace the sprite will throw an error
    __jo_sprite_pic[id].color_mode=COL_16;
    texture->size = JO_MULT_BY_32(sky.width & 0x1f8) | texture->height;

    set_palette(sky_PALETTE, id);
}

/******************************
Load binary image
CAUTION : Jo Engine might throw an error if the texture's width/4 < 8,
The alternative would be to divide the height instead, but since the height doesn't need to be a multiple of 8 (ex.:3),
dividing by 4 might break everything (Ex.: 3/4 = 0).
Turn off the debug options in your makefile if that's an issue
******************************/
void			    load_4bits_bin_image(char * name, unsigned short width, unsigned short height)
{
    jo_img *  img;
    jo_texture_definition   *texture;
    void * currentAddr = (void*)LWRAM;  //Makes it easier to track the address

    /**CD functions**/
    jo_file file;
    jo_fs_open(&file, name);
    jo_fs_read_next_bytes(&file, (void*)LWRAM, (8+(width*height>>1)+32));  //8 bytes for width, height and pointer, then the width*heigth/2 (since we use 4 bits data) + the size of the palette (always 16*sizeof(Uint16))
    jo_fs_close(&file);  //or use this : GFS_Load(GFS_NameToId(name), 0, (void *)LWRAM, (4+(width*height>>1)+32));

    /**Set the pointers**/
    img = (jo_img*)currentAddr;               currentAddr=(void*)(currentAddr+sizeof(jo_img));
    img->width=img->width>>2;  //Quick and dirty way to get around the fact that Jo Engine doesn't support 4 bits pixel data.
    img->data = (jo_color*)(currentAddr);      currentAddr=(void*)(currentAddr+(sizeof(Uint16)*img->width * img->height));

    /**Adds the img in memory**/
    int id = jo_sprite_add(img);

    /**Here we just fix the issue caused by using the add 8 bits image for 4 bits images**/
    texture=&__jo_sprite_def[id];
    texture->width=img->width<<2;  //Ghetto technique for compatibility with Jo Engine, but trying to replace the sprite will throw an error
    __jo_sprite_pic[id].color_mode=COL_16;
    texture->size = JO_MULT_BY_32(texture->width & 0x1f8) | texture->height;

    /**Here we set the palette**/
    set_palette((unsigned short *)currentAddr, id);
}



void			    jo_main(void)
{
	jo_core_init(JO_COLOR_Black);
	load_4bits_image();

	/**Note that loading each file this way is slow, use a server function instead and load several files (async loading)**/
	load_4bits_bin_image("B1.BIN", 48, 48);
	load_4bits_bin_image("BOX.BIN", 24, 24);
	load_4bits_bin_image("SONIC.BIN", 128, 128); /**Auto-compressed image, which is why it doesn't look optimal**/

	jo_core_add_callback(my_draw);
	jo_core_run();
}

/*
** END OF FILE
*/
